/*  -*- C++ -*-
 * lextable.C
 * @(#) implementation for the hash table used by the Murphi lexer.
 *
 * Copyright (C) 1992 - 1999 by the Board of Trustees of
 * Leland Stanford Junior University.
 *
 * License to use, copy, modify, sell and/or distribute this software
 * and its documentation any purpose is hereby granted without royalty,
 * subject to the following terms and conditions:
 *
 * 1.  The above copyright notice and this permission notice must
 * appear in all copies of the software and related documentation.
 *
 * 2.  The name of Stanford University may not be used in advertising or
 * publicity pertaining to distribution of the software without the
 * specific, prior written permission of Stanford.
 *
 * 3.  This software may not be called "Murphi" if it has been modified
 * in any way, without the specific prior written permission of David L.
 * Dill.
 *
 * 4.  THE SOFTWARE IS PROVIDED "AS-IS" AND STANFORD MAKES NO
 * REPRESENTATIONS OR WARRANTIES, EXPRESS OR IMPLIED, BY WAY OF EXAMPLE,
 * BUT NOT LIMITATION.  STANFORD MAKES NO REPRESENTATIONS OR WARRANTIES
 * OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR PURPOSE OR THAT THE
 * USE OF THE SOFTWARE WILL NOT INFRINGE ANY PATENTS, COPYRIGHTS
 * TRADEMARKS OR OTHER RIGHTS. STANFORD SHALL NOT BE LIABLE FOR ANY
 * LIABILITY OR DAMAGES WITH RESPECT TO ANY CLAIM BY LICENSEE OR ANY
 * THIRD PARTY ON ACCOUNT OF, OR ARISING FROM THE LICENSE, OR ANY
 * SUBLICENSE OR USE OF THE SOFTWARE OR ANY SERVICE OR SUPPORT.
 *
 * LICENSEE shall indemnify, hold harmless and defend STANFORD and its
 * trustees, officers, employees, students and agents against any and all
 * claims arising out of the exercise of any rights under this Agreement,
 * including, without limiting the generality of the foregoing, against
 * any damages, losses or liabilities whatsoever with respect to death or
 * injury to person or damage to property arising from or out of the
 * possession, use, or operation of Software or Licensed Program(s) by
 * LICENSEE or its customers.
 *
 * Read the file "license" distributed with these sources, or call
 * Murphi with the -l switch for additional information.
 * 
 */

/* 
 * Original Author: Ralph Melton
 * 
 * Update:
 *
 * C. Norris Ip 
 * Subject: Case insensitive keywords and case sensitive identifiers
 * Modified: April 93
 * 
 * Seungjoon Park
 * Subject: liveness syntax
 * Last modified: Novemeber  5, 93
 *
 * C. Norris Ip 
 * Subject: symmetry syntax
 * Last modified: November 8, 93
 *
 * C. Norris Ip
 * Subject: Extension for {multiset, undefined value, general union}
 * First modified: May 24, 94
 *
 */ 

/* 
 * Cached Murphi Release 4.2
 * Author: G. Della Penna
 * Released: 01 May 2003
 * Notes: Cached Murphi now supports (finite precision) real numbers.
 *       
 */

#include "mu.h"
#include <string.h>
#include <ctype.h>

/********************
  Hash table
  -- Mostly copied from the hash table used in CS143, Spring '92. rlm
  -- A simple and mediocre hash table implementation.
  -- This will go into an infinite loop when the hash table fills up.
  ********************/

/********************
  Constant
  -- constant for int lextable::hash(char *str) const
  ********************/
#define M1		71
#define M2		578393
#define M3		358723

/********************
  class lexid
  -- from mu.h
  ********************/
lexid::lexid(char *name, int lextype)
{
  this->name = new char[ strlen(name) + 1]; // +1 for the '\0'.
  strcpy(this->name,name);
  this->lextype = lextype;
}

/********************
  utilities
  ********************/
char *my_strtolower(char * str)
{
  char *ret;

  ret = new char[ strlen(str) + 1];
  strcpy(ret, str);
  for (int i=0; i<strlen(ret); i++)
    ret[i]=tolower(ret[i]);
  return ret;
}

/********************
  class lextable
  ********************/
int lextable::hash(char *str) const
/* This returns an integer hash value for a string. */
{
  char c;
  int h = 0;
  while ((c = *str++) != '\0')
    h = (h * M1 + (int) c );
  return(h);
}

int lextable::rehash(int h) const
/* If hash location is already full, this computes a new hash address to 
   try. */
{
  return(h +1);
}

bool lextable::reserved(char *str)
{
  // in order to allow at least and at most one entry
  // of reserved words in the main lex table,
  // the lower case version is considered not reserved here,
  // so that the same enter routine can be used in the 
  // main lex table for initializing keywords.

  int h;
  char *lowerstr;

  lowerstr = my_strtolower(str);
  h = hash(lowerstr);

  while (1) {
    int i = (h & RESERVETABLESIZE-1);
    lexid *entry = rtable[i];
    if (entry == NULL) {
      return(FALSE);
    }
    else if ( strcmp(entry->getname(), lowerstr) == 0) {
      return(TRUE);
    }
    else {
      /* try again */
      h = rehash(h);
    }
  }
}

lexid *lextable::enter_reserved(char *str, int lextype )
{
  int h;
  char *lowerstr;

  lowerstr = my_strtolower(str);
  h = hash(lowerstr);

  while (1) {
    int i = (h & RESERVETABLESIZE-1);
    lexid *entry = rtable[i];
    if (entry == NULL) {
      /* enter it and return */
      rtable[i] = entry = new lexid(lowerstr, lextype);
      return(entry);
    }
    else if ( strcmp(entry->getname(), lowerstr) == 0) {
      return(entry);
    }
    else {
      /* try again */
      h = rehash(h);
    }
  }
}

lexid *lextable::enter(char *str, int lextype )
{
  int h = hash(str);
  while (1) {
    int i = (h & LEXTABLESIZE-1);
    lexid *entry = table[i];
    if (entry == NULL) {
      // there is no entry in the lex table
      // but is it a reserved word in a different case context?
      if (reserved(str))
      {
	// yes, it is.  get the entry in the reserved words table
        return enter_reserved(str);
      }
      else
      {
        /* no, it is not.  Enter it and return */
        table[i] = entry = new lexid(str, lextype);
        return(entry);
      }
    }
    else if ( strcmp(entry->getname(), str) == 0) {
      return(entry);
    }
    else {
      /* try again */
      h = rehash(h);
    }
  }
}

lextable::lextable()
{
  int i; /* loop index. */
  for( i=0; i < LEXTABLESIZE; i++) table[i] = NULL;

  // table[] is case sensitive and rtable[] is case insensitive
  // before entering new entry, the string is checked again
  // the reserve table in lower case.  Therefore, all reserved words
  // are case insensitive and everything else are case sensitive
  enter_reserved(STR2PTR("end"), END);
  enter_reserved(STR2PTR("program"), PROGRAM); 
  // enter_reserved(STR2PTR("process"), PROCESS); 
  enter_reserved(STR2PTR("procedure"), PROCEDURE);
  enter_reserved(STR2PTR("endprocedure"), ENDPROCEDURE);
  enter_reserved(STR2PTR("function"), FUNCTION);
  enter_reserved(STR2PTR("endfunction"), ENDFUNCTION);
  enter_reserved(STR2PTR("rule"), RULE);
  enter_reserved(STR2PTR("endrule"), ENDRULE);
  enter_reserved(STR2PTR("ruleset"), RULESET);
  enter_reserved(STR2PTR("endruleset"), ENDRULESET);
  enter_reserved(STR2PTR("alias"), ALIAS);
  enter_reserved(STR2PTR("endalias"), ENDALIAS);
  enter_reserved(STR2PTR("if"), IF);
  enter_reserved(STR2PTR("then"), THEN);
  enter_reserved(STR2PTR("elsif"), ELSIF);
  enter_reserved(STR2PTR("else"), ELSE);
  enter_reserved(STR2PTR("endif"), ENDIF);
  enter_reserved(STR2PTR("switch"), SWITCH);
  enter_reserved(STR2PTR("case"), CASE);
  enter_reserved(STR2PTR("endswitch"), ENDSWITCH);
  enter_reserved(STR2PTR("for"), FOR);
  enter_reserved(STR2PTR("forall"), FORALL);
  enter_reserved(STR2PTR("exists"), EXISTS);
  enter_reserved(STR2PTR("in"), IN);
  enter_reserved(STR2PTR("do"), DO);
  enter_reserved(STR2PTR("endfor"), ENDFOR);
  enter_reserved(STR2PTR("endforall"), ENDFORALL);
  enter_reserved(STR2PTR("endexists"), ENDEXISTS);
  enter_reserved(STR2PTR("while"), WHILE);
  enter_reserved(STR2PTR("endwhile"), ENDWHILE);
  enter_reserved(STR2PTR("return"), RETURN);
  enter_reserved(STR2PTR("to"),  TO);
  enter_reserved(STR2PTR("begin"), bEGIN);
  enter_reserved(STR2PTR("by"), BY);
  enter_reserved(STR2PTR("clear"), CLEAR);
  enter_reserved(STR2PTR("error"), ERROR);
  enter_reserved(STR2PTR("assert"), ASSERT);
  enter_reserved(STR2PTR("put"), PUT);
  enter_reserved(STR2PTR("const"), CONST);
  enter_reserved(STR2PTR("type"), TYPE);
  enter_reserved(STR2PTR("var"), VAR);
  enter_reserved(STR2PTR("enum"), ENUM);
  enter_reserved(STR2PTR("interleaved"), INTERLEAVED);
  enter_reserved(STR2PTR("record"), RECORD);
  enter_reserved(STR2PTR("array"), ARRAY);
  enter_reserved(STR2PTR("of"), OF);
  enter_reserved(STR2PTR("endrecord"), ENDRECORD);
  enter_reserved(STR2PTR("startstate"), STARTSTATE);
  enter_reserved(STR2PTR("endstartstate"), ENDSTARTSTATE);
  enter_reserved(STR2PTR("invariant"),  INVARIANT);
  enter_reserved(STR2PTR("traceuntil"), TRACEUNTIL);

  /* liveness */
  enter_reserved(STR2PTR("fairness"),  FAIRNESS);
  enter_reserved(STR2PTR("fairnessset"), FAIRNESSSET);
  enter_reserved(STR2PTR("endfairnessset"), ENDFAIRNESSSET);
  enter_reserved(STR2PTR("liveness"),  LIVENESS);
  enter_reserved(STR2PTR("livenessset"), LIVENESSSET);
  enter_reserved(STR2PTR("endlivenessset"), ENDLIVENESSSET);
  enter_reserved(STR2PTR("always"),  ALWAYS);
  enter_reserved(STR2PTR("eventually"),  EVENTUALLY);
  enter_reserved(STR2PTR("until"),  UNTIL);
  enter_reserved(STR2PTR("unfair"),  UNFAIR);

  /* scalarset */
  enter_reserved(STR2PTR("scalarset"), SCALARSET);
  enter_reserved(STR2PTR("ismember"), ISMEMBER);

  /* undefined */
  enter_reserved(STR2PTR("undefine"), UNDEFINE);
  enter_reserved(STR2PTR("isundefined"), ISUNDEFINED);
  enter_reserved(STR2PTR("undefined"), UNDEFINED);

  /* general union */
  enter_reserved(STR2PTR("union"), UNION);

  /* multiset */
  enter_reserved(STR2PTR("multiset"), MULTISET);
  enter_reserved(STR2PTR("multisetremove"), MULTISETREMOVE);
  enter_reserved(STR2PTR("multisetremovepred"), MULTISETREMOVEPRED);
  enter_reserved(STR2PTR("multisetadd"), MULTISETADD);
  enter_reserved(STR2PTR("multisetcount"), MULTISETCOUNT);
  enter_reserved(STR2PTR("choose"), CHOOSE);
  enter_reserved(STR2PTR("endchoose"), ENDCHOOSE);

  /* first definitions. */
  enter_reserved(STR2PTR("boolean"), ID);
  enter_reserved(STR2PTR("true"), ID);
  enter_reserved(STR2PTR("false"), ID);

  /* AP: added some reserved words */
  enter_reserved(STR2PTR("real"), REAL);
  enter_reserved(STR2PTR("log"), LOG);
  enter_reserved(STR2PTR("log10"), LOG10);
  enter_reserved(STR2PTR("exp"), EXP);
  enter_reserved(STR2PTR("sin"), SIN);
  enter_reserved(STR2PTR("cos"), COS);
  enter_reserved(STR2PTR("tan"), TAN);
  enter_reserved(STR2PTR("fabs"), FABS);
  enter_reserved(STR2PTR("floor"), FLOOR);
  enter_reserved(STR2PTR("ceil"), CEIL);
  enter_reserved(STR2PTR("sqrt"), SQRT);
  enter_reserved(STR2PTR("fmod"), FMOD);
  enter_reserved(STR2PTR("pow"), POW);
  enter_reserved(STR2PTR("asin"), ASIN);
  enter_reserved(STR2PTR("acos"), ACOS);
  enter_reserved(STR2PTR("atan"), ATAN);
  enter_reserved(STR2PTR("sinh"), SINH);
  enter_reserved(STR2PTR("cosh"), COSH);
  enter_reserved(STR2PTR("tanh"), TANH);
}

/********************
  variable declarations
  ********************/
lextable ltable;    

/****************************************
  * 8 March 94 Norris Ip:
  merge with the latest rel2.6
  ******
  * Subject: Extension for {multiset, undefined value, general union}
  * 24 May 94 Norris Ip:
  added keyword undefined
  * 25 May 94 Norris Ip:
  changed keyword scalarsetunion to union
  * 7 OCT 94 Norris Ip:
  added keywords on multiset
****************************************/

/********************
 $Log: lextable.C,v $
 Revision 2.1  2003/05/01 18:13:05  giuseppe
 real numbers extension - final release

 Revision 1.2  1999/01/29 07:49:13  uli
 bugfixes

 Revision 1.4  1996/08/07 18:54:00  ip
 last bug fix on NextRule/SetNextEnabledRule has a bug; fixed this turn

 Revision 1.3  1996/08/07 00:59:13  ip
 Fixed bug on what_rule setting during guard evaluation; otherwise, bad diagnoistic message on undefine error on guard

 Revision 1.2  1996/08/06 23:57:39  ip
 fixed while code generation bug

 Revision 1.1  1996/08/06 23:56:55  ip
 Initial revision

 ********************/
